@@ -80,6 +80,7 @@ static const R_CallMethodDef callMethods[] =
     {"git2r_branch_set_upstream", (DL_FUNC)&git2r_branch_set_upstream, 2},
     {"git2r_branch_target", (DL_FUNC)&git2r_branch_target, 1},
     {"git2r_branch_upstream_canonical_name", (DL_FUNC)&git2r_branch_upstream_canonical_name, 1},
+    {"git2r_checkout_path", (DL_FUNC)&git2r_checkout_path, 2},
     {"git2r_checkout_tree", (DL_FUNC)&git2r_checkout_tree, 3},
     {"git2r_clone", (DL_FUNC)&git2r_clone, 6},
     {"git2r_commit", (DL_FUNC)&git2r_commit, 4},
@@ -117,9 +118,11 @@ static const R_CallMethodDef callMethods[] =
     {"git2r_remote_list", (DL_FUNC)&git2r_remote_list, 1},
     {"git2r_remote_remove", (DL_FUNC)&git2r_remote_remove, 2},
     {"git2r_remote_rename", (DL_FUNC)&git2r_remote_rename, 3},
+    {"git2r_remote_set_url", (DL_FUNC)&git2r_remote_set_url, 3},
     {"git2r_remote_url", (DL_FUNC)&git2r_remote_url, 2},
+    {"git2r_remote_ls", (DL_FUNC)&git2r_remote_ls, 3},
     {"git2r_repository_can_open", (DL_FUNC)&git2r_repository_can_open, 1},
-    {"git2r_repository_discover", (DL_FUNC)&git2r_repository_discover, 1},
+    {"git2r_repository_discover", (DL_FUNC)&git2r_repository_discover, 2},
     {"git2r_repository_fetch_heads", (DL_FUNC)&git2r_repository_fetch_heads, 1},
     {"git2r_repository_head", (DL_FUNC)&git2r_repository_head, 1},
     {"git2r_repository_head_detached", (DL_FUNC)&git2r_repository_head_detached, 1},
@@ -135,6 +138,7 @@ static const R_CallMethodDef callMethods[] =
     {"git2r_revwalk_contributions", (DL_FUNC)&git2r_revwalk_contributions, 4},
     {"git2r_revwalk_list", (DL_FUNC)&git2r_revwalk_list, 5},
     {"git2r_signature_default", (DL_FUNC)&git2r_signature_default, 1},
+    {"git2r_ssl_cert_locations", (DL_FUNC)&git2r_ssl_cert_locations, 2},
     {"git2r_stash_drop", (DL_FUNC)&git2r_stash_drop, 2},
     {"git2r_stash_list", (DL_FUNC)&git2r_stash_list, 1},
     {"git2r_stash_save", (DL_FUNC)&git2r_stash_save, 6},
