@@ -51,16 +51,16 @@ SEXP git2r_stash_drop(SEXP repo, SEXP index)
     git_repository *repository = NULL;
 
     if (git2r_arg_check_integer_gte_zero(index))
-        git2r_error(git2r_err_integer_gte_zero_arg, __func__, "index");
+        git2r_error(__func__, NULL, "'index'", git2r_err_integer_gte_zero_arg);
 
     repository = git2r_repository_open(repo);
     if (!repository)
-        git2r_error(git2r_err_invalid_repository, __func__, NULL);
+        git2r_error(__func__, NULL, git2r_err_invalid_repository, NULL);
 
     err = git_stash_drop(repository, INTEGER(index)[0]);
     git_repository_free(repository);
-    if (GIT_OK != err)
-        git2r_error(git2r_err_from_libgit2, __func__, giterr_last()->message);
+    if (err)
+        git2r_error(__func__, giterr_last(), NULL, NULL);
 
     return R_NilValue;
 }
@@ -84,7 +84,7 @@ int git2r_stash_init(
     git_commit *commit = NULL;
 
     err = git_commit_lookup(&commit, repository, source);
-    if (GIT_OK != err)
+    if (err)
         return err;
     git2r_commit_init(commit, repo, dest);
     git_commit_free(commit);
@@ -127,7 +127,7 @@ static int git2r_stash_list_cb(
             cb_data->repository,
             cb_data->repo,
             stash);
-        if (GIT_OK != err)
+        if (err)
             return err;
     }
 
@@ -151,11 +151,11 @@ SEXP git2r_stash_list(SEXP repo)
 
     repository = git2r_repository_open(repo);
     if (!repository)
-        git2r_error(git2r_err_invalid_repository, __func__, NULL);
+        git2r_error(__func__, NULL, git2r_err_invalid_repository, NULL);
 
     /* Count number of stashes before creating the list */
     err = git_stash_foreach(repository, &git2r_stash_list_cb, &cb_data);
-    if (GIT_OK != err)
+    if (err)
         goto cleanup;
 
     PROTECT(list = allocVector(VECSXP, cb_data.n));
@@ -172,8 +172,8 @@ cleanup:
     if (R_NilValue != list)
         UNPROTECT(1);
 
-    if (GIT_OK != err)
-        git2r_error(git2r_err_from_libgit2, __func__, giterr_last()->message);
+    if (err)
+        git2r_error(__func__, giterr_last(), NULL, NULL);
 
     return list;
 }
@@ -209,17 +209,17 @@ SEXP git2r_stash_save(
     git_signature *c_stasher = NULL;
 
     if (git2r_arg_check_logical(index))
-        git2r_error(git2r_err_logical_arg, __func__, "index");
+        git2r_error(__func__, NULL, "'index'", git2r_err_logical_arg);
     if (git2r_arg_check_logical(untracked))
-        git2r_error(git2r_err_logical_arg, __func__, "untracked");
+        git2r_error(__func__, NULL, "'untracked'", git2r_err_logical_arg);
     if (git2r_arg_check_logical(ignored))
-        git2r_error(git2r_err_logical_arg, __func__, "ignored");
+        git2r_error(__func__, NULL, "'ignored'", git2r_err_logical_arg);
     if (git2r_arg_check_signature(stasher))
-        git2r_error(git2r_err_signature_arg, __func__, "stasher");
+        git2r_error(__func__, NULL, "'stasher'", git2r_err_signature_arg);
 
     repository = git2r_repository_open(repo);
     if (!repository)
-        git2r_error(git2r_err_invalid_repository, __func__, NULL);
+        git2r_error(__func__, NULL, git2r_err_invalid_repository, NULL);
 
     if (LOGICAL(index)[0])
         flags |= GIT_STASH_KEEP_INDEX;
@@ -229,7 +229,7 @@ SEXP git2r_stash_save(
         flags |= GIT_STASH_INCLUDE_IGNORED;
 
     err = git2r_signature_from_arg(&c_stasher, stasher);
-    if (GIT_OK != err)
+    if (err)
         goto cleanup;
 
     err = git_stash_save(
@@ -238,7 +238,7 @@ SEXP git2r_stash_save(
         c_stasher,
         CHAR(STRING_ELT(message, 0)),
         flags);
-    if (GIT_OK != err) {
+    if (err) {
         if (GIT_ENOTFOUND == err)
             err = GIT_OK;
         goto cleanup;
@@ -260,8 +260,8 @@ cleanup:
     if (R_NilValue != result)
         UNPROTECT(1);
 
-    if (GIT_OK != err)
-        git2r_error(git2r_err_from_libgit2, __func__, giterr_last()->message);
+    if (err)
+        git2r_error(__func__, giterr_last(), NULL, NULL);
 
     return result;
 }
